import pytest
from xchainpy_bitcoincash.client import Client
from xchainpy_util.asset import Asset


class TestBitcoincashClient:
    memo = 'SWAP:THOR.RUNE'
    phrase = 'atom green various power must another rent imitate gadget creek fat then'
    testnet_address = 'bchtest:qpd7jmj0hltgxux06v9d9u6933vq7zd0kyjlapya0g'
    mainnet_address = 'bitcoincash:qp4kjpk684c3d9qjk5a37vl2xn86wxl0f5j2ru0daj'

    bch_asset = Asset('BCH', 'BCH')
    
    @pytest.fixture
    def client(self):
        self.client = Client(phrase=self.phrase, network='testnet')
        yield
        self.client.purge_client()

    def test_set_phrase_should_return_correct_address(self, client):
        self.client.set_network('testnet')
        assert self.client.set_phrase(self.phrase) == self.testnet_address

        self.client.set_network('mainnet')
        assert self.client.set_phrase(self.phrase) == self.mainnet_address


    def test_invalid_phrase(self):
        with pytest.raises(Exception) as err:
            assert Client(phrase='Invalid Phrase')
        assert str(err.value) == "Invalid Phrase"

    def test_right_phrase(self, client):
        assert self.client.set_phrase(self.phrase) == self.testnet_address

    def test_validate_address(self, client):
        assert self.client.validate_address(address=self.mainnet_address)
        assert self.client.validate_address(address=self.testnet_address)

    @pytest.mark.asyncio
    async def test_has_balances(self, client):
        balance = await self.client.get_balance()
        assert balance.asset == self.bch_asset
        assert isinstance(balance.amount, float)

    @pytest.mark.asyncio
    async def test_get_transaction_with_hash(self, client):
        tx_data = await self.client.get_transaction_data('0d5764c89d3fbf8bea9b329ad5e0ddb6047e72313c0f7b54dcb14f4d242da64b')
        assert tx_data.tx_hash == '0d5764c89d3fbf8bea9b329ad5e0ddb6047e72313c0f7b54dcb14f4d242da64b'
        assert len(tx_data.tx_from) == 1
        assert tx_data.tx_from[0].address == 'bchtest:qzyrvsm6z4ucrhaq4zza3wylre7mavvldgr67jrxt4'
        assert str(tx_data.tx_from[0].amount) == '0.04008203'

        assert len(tx_data.tx_to) == 2
        assert tx_data.tx_to[0].address == 'bchtest:qq235k7k9y5cwf3s2vfpxwgu8c5497sxnsdnxv6upc'
        assert str(tx_data.tx_to[0].amount) == '0.04005704'


    @pytest.mark.asyncio
    async def test_transfer_with_memo_and_fee_rate(self, client):
        fee_rates = await self.client.get_fee_rates()
        fee_rate = fee_rates['fast']
        balance = await self.client.get_balance()
        if balance.amount > 0:
            amount = 0.00000001
            tx_id = await self.client.transfer(amount, 'bchtest:qzt6sz836wdwscld0pgq2prcpck2pssmwge9q87pe9', self.memo, fee_rate)
            assert tx_id
            
    @pytest.mark.asyncio
    async def test_fee_and_rates_normal_tx(self, client):
        fees_and_rates = await self.client.get_fees_with_rates()
        fees = fees_and_rates['fees']
        rates = fees_and_rates['rates']
        assert fees['fastest']
        assert fees['fast']
        assert fees['average']
        assert rates['fastest']
        assert rates['fast']
        assert rates['average']

    @pytest.mark.asyncio
    async def test_fee_and_rates_with_memo(self, client):
        fees_and_rates = await self.client.get_fees_with_rates(self.memo)
        fees = fees_and_rates['fees']
        rates = fees_and_rates['rates']
        assert fees['fastest']
        assert fees['fast']
        assert fees['average']
        assert rates['fastest']
        assert rates['fast']
        assert rates['average']

    @pytest.mark.asyncio
    async def test_estimated_fees_normal_tx(self, client):
        fees = await self.client.get_fees()
        assert fees['fastest']
        assert fees['fast']
        assert fees['average']

    @pytest.mark.asyncio
    async def test_different_fees_normal_tx(self, client):
        fees = await self.client.get_fees()

        assert fees['fastest'] > fees['fast']
        assert fees['fast'] > fees['average']


from xchainpy_util.asset import Asset
from xchainpy_client.models.balance import Balance
from . import haskoin_api
from bitcash.wallet import Key, PrivateKeyTestnet
from cashaddress import convert
from cashaddress.convert import Address
from .crypto import mnemonic_to_private_key, private_key_to_address
from mnemonic.mnemonic import Mnemonic
from xchainpy_crypto.crypto import validate_phrase
from . import utils
from xchainpy_client.interface import IXChainClient
from bitcash import transaction, PrivateKey, PrivateKeyTestnet, network

class IBitcoinCashClient():
    def derive_path(self) -> str :
        pass
    def get_fees_with_rates(self , memo = None):
        pass
    def get_fees_with_memo(self):
        pass
    def get_fee_rates(self):
        pass

class Client(IBitcoinCashClient , IXChainClient):
    def __init__(self, phrase:str, network='testnet', client_url:utils.ClientUrl=None):
        self.set_network(network)
        self.client_url = client_url if client_url else self.get_default_client_url()
        self.set_phrase(phrase)

    def get_default_client_url(self) -> utils.ClientUrl:
        """Get default ClientURL based on the network

        :returns: the transaction hash
        """
        return utils.ClientUrl("https://api.haskoin.com/bchtest" ,"https://api.haskoin.com/bch")

    def set_phrase(self, phrase: str):
        """Set/update a new phrase

        :param phrase: A new phrase
        :type phrase: str
        :returns: The address from the given phrase
        :raises: 'Invalid Phrase' if the given phrase is invalid
        """
        if validate_phrase(phrase):
            self.phrase = phrase
            self.address = self.get_address()
            return self.address
        else:
            raise Exception("Invalid Phrase")

    def get_address(self):
        """Get the current address

        :returns: The current address
        :raises: "Phrase must be provided" if phrase has not been set before
        :raises: "Address not defined" if failed creating account from phrase
        """
        if self.phrase:
            try:
                priv_key = self.get_private_key(self.phrase)
                address = private_key_to_address(priv_key, self.network)
                return str(address)
            except:
                raise Exception("Address not defined")
        else:
            raise Exception("Phrase must be provided")

    def get_private_key(self , phrase:str=None):
        """Get private key

        :param phrase: The phrase to be used for generating privkey
        :type phrase: str
        :returns: The privkey generated from the given phrase
        """
        try:
            privKey = mnemonic_to_private_key(phrase or self.phrase, self.network)
            return privKey
        except:
            raise Exception("Invalid Phrase")


    def derive_path(self):
        """Get DerivePath

        :returns: The bitcoin cash derivation path based on the network
        """
        return utils.get_derive_path().testnet if self.network == 'testnet' else utils.get_derive_path().mainnet

    def purge_client(self):
        """Purge client
        """
        self.phrase = ''

    def get_network(self):
        """Get the current network

        :returns: The current network. (mainnet or testnet)
        """
        return self.network

    def set_network(self, network: str):
        """Set/update the current network

        :param network: mainnet or testnet
        :type network: str
        :raises: Network must be provided if network has not been set before
        :raises: `Invalid network' if the given network is invalid
        """
        if not network:
            raise Exception("Network must be provided")
        if not network in ['testnet', 'mainnet']:
                raise Exception('Invalid network')
        else:
            self.network = network

    def set_client_url(self , url : utils.ClientUrl):
        """Set/Update the node url

        :param url: The new node url
        :type url: str
        """
        self.client_url = url

    def get_client_url(self):
        """Get the client url

        :returns: The client url based on the current network
        """
        return self.get_client_url_by_network(self.get_network())

    def get_client_url_by_network(self, network):
        """Get the client url by network

        :param network: testnet or mainnet
        :type network: str
        :returns: The client url based on the network
        """
        if network == "testnet":
            return self.client_url.testnet
        if network == "mainnet":
            return self.client_url.mainnet
        raise Exception('invalid network')

    def validate_address(self, address: str):
        """Validate the given address

        :param address: address
        :type address: str
        :returns: true or false
        """
        return convert.is_valid(address)

    async def get_balance(self, address:str=""):
        """Get the BCH balance of a given address

        :param address: By default, it will return the balance of the current wallet. (optional)
        :type address: str
        :returns: The BCH balance of the address
        :raises: "Invalid address" if the given address is an invalid address
        """
        try:            
            account = await haskoin_api.get_account(self.get_client_url() , address or self.get_address())
            if not account:
                raise Exception("Invalid Address")
            
            balance = Balance(Asset.from_str("BCH.BCH"), account.confirmed * 10 ** -8)
            return balance
                
        except Exception as err:
            raise Exception(str(err))

    async def get_transaction_data(self, tx_id: str):
        """Get the transaction details of a given transaction id

        :param tx_id: The transaction id
        :type str: str
        :returns: The transaction details of the given transaction id
        :raises: "Invalid address" if the given transaction id is an invalid one
        """
        try:
            tx = await haskoin_api.get_transaction(self.get_client_url() , tx_id)

            if not tx:
                raise Exception("Invalid tx_id")

            data = utils.parse_tx(tx)
            return data

        except Exception as err:
            raise Exception(str(err))

    async def get_fees_with_rates(self, memo: str = ''):
        """Get the rates and fees

        :param memo: The memo to be used for fee calculation (optional)
        :type memo: str
        :returns: The fees and rates
        """
        next_block_fee_rates = await haskoin_api.get_suggested_tx_fee()

        rates = {
            'fastest': next_block_fee_rates * 5,
            'fast': next_block_fee_rates * 1,
            'average': next_block_fee_rates * 0.5
        }
        fees = {
            'fastest': utils.calc_fee(rates['fastest'], memo),
            'fast': utils.calc_fee(rates['fast'], memo),
            'average': utils.calc_fee(rates['average'], memo)
        }
        return {
            'rates': rates,
            'fees': fees
        }

    async def get_fees(self):
        """Get the current fees

        :returns: The fees without memo
        """
        try:
            fees = (await self.get_fees_with_rates())['fees']
            return fees
        except Exception as err:
            raise Exception(str(err))

    async def get_fees_with_memo(self, memo: str):
        """Get the fees for transactions with memo
        If you want to get `fees` and `fee_rates` at once, use `get_fees_with_rates` method

        :param memo: The memo to be used for fee calculation (optional)
        :type memo: str
        :returns: The fees with memo
        """
        try:
            fees = (await self.get_fees_with_rates(memo))['fees']
            return fees
        except Exception as err:
            raise Exception(str(err))

    async def get_fee_rates(self):
        """Get the fee rates for transactions without a memo
        If you want to get `fees` and `fee_rates` at once, use `get_fees_with_rates` method

        :returns: The fee rate
        """
        try:
            rates = (await self.get_fees_with_rates())['rates']
            return rates
        except Exception as err:
            raise Exception(str(err))
    

    async def transfer(self, amount, recipient, memo: str = None, fee_rate=None) -> str:
        """Transfer BCH

        :param amount: amount of BTC to transfer (don't multiply by 10**8)
        :type amount: int, float, decimal
        :param recipient: destination address
        :type recipient: str
        :param memo: optional memo for transaction
        :type memo: str
        :param fee_rate: fee rates for transaction
        :type fee_rate: int
        :returns: The transaction hash
        """
        try:
            fee_rate = fee_rate or (await self.get_fee_rates())['fast']

            KeyClass = PrivateKeyTestnet if self.get_network() == "testnet" else PrivateKey
            key = KeyClass.from_hex(self.get_private_key())

            tx_hex = await utils.build_tx(amount , recipient , memo , fee_rate , self.get_address(),self.get_network(), self.get_client_url(), key)
            
            network.NetworkAPI.broadcast_tx_testnet(tx_hex)
            return transaction.calc_txid(tx_hex)
            

        except Exception as err:
            raise Exception(str(err))


import datetime
import bitcash

from cashaddress import convert
from . import haskoin_api

from bitcash import transaction, PrivateKey, PrivateKeyTestnet
from .models.api_types import Transaction, TxUnspent
from xchainpy_util.asset import Asset
from xchainpy_util.chain import BITCOINCASH
from xchainpy_client.models import tx_types


BCH_DECIMAL = 8
DEFAULT_SUGGESTED_TRANSACTION_FEE = 1

class DerivePath:
    def __init__(self, index:int=0):

        self._mainnet = f"m/44'/145'/0'/0/{index}"
        self._testnet = f"m/44'/1'/0'/0/{index}"

    @property
    def mainnet(self):
        return self._mainnet

    @mainnet.setter
    def mainnet(self, mainnet):
        self._mainnet = mainnet

    @property
    def testnet(self):
        return self._testnet

    @testnet.setter
    def testnet(self, testnet):
        self._testnet = testnet

def get_derive_path(index: int = 0):
    return DerivePath(index=index)


class ClientUrl:
    def __init__(self , testnet , mainnet):
        self._testnet : str = testnet
        self._mainnet : str = mainnet

    @property
    def mainnet(self):
        return self._mainnet

    @mainnet.setter
    def mainnet(self, mainnet : str):
        self._mainnet = mainnet

    @property
    def testnet(self):
        return self._testnet

    @testnet.setter
    def testnet(self, testnet : str):
        self._testnet = testnet
    

def parse_tx(tx : Transaction):
    """Parse tx
    :param tx: The transaction to be parsed
    :type tx: str
    :returns: The transaction parsed from the binance tx
    """
    asset = Asset.from_str(f'{BITCOINCASH}.BCH')
    tx_from = [tx_types.TxFrom(i.address, i.value * 10 ** -8) for i in tx.inputs]
    tx_to = [tx_types.TxTo(i.address, i.value * 10 ** -8) for i in tx.outputs]
    tx_date = datetime.datetime.fromtimestamp(tx.time)
    tx_type = 'transfer'
    tx_hash = tx.txid

    tx = tx_types.TX(asset, tx_from, tx_to, tx_date, tx_type, tx_hash)
    return tx

def calc_fee(fee_rate , memo=None , utxos=None):
    """Calculate fees based on fee rate and memo

    :param fee_rate: fee rates
    :type fee_rate: str
    :param memo: memo (optional)
    :type memo: str
    :param utxos: A list of utxos (optional)
    :type utxos: str
    :returns: The calculated fees based on fee rate and the memo
    """
    # key = PrivateKey()
    # random_address = key.address

    if utxos:
        utxos_len = len(utxos)
    else:
        utxos_len = 0

    # outputs = [(random_address, 0, 'bch')]
    # for i, output in enumerate(outputs):
    #     dest, amount, currency = output
    #     outputs[i] = (dest, transaction.currency_to_satoshi_cached(amount, currency))
    # num_outputs = len(outputs) + 1
    num_outputs = 2

    total_op_return_size = 0
    if memo:
        message_chunks = transaction.chunk_data(memo, transaction.MESSAGE_LIMIT)
        for message in message_chunks:
            total_op_return_size += transaction.get_op_return_size(message, custom_pushdata=False)

    calculated_fee = transaction.estimate_tx_fee(utxos_len, num_outputs, fee_rate, False, total_op_return_size)
    return calculated_fee

async def scan_UTXOs(client_url, address):
    """Scan UTXOs from sochain

    :param network: testnet or mainnet
    :type network: str
    :param address: address
    :type address: str
    :returns: The UTXOs of the given address
    """
    unspents = await haskoin_api.get_unspent_transactions(client_url, address)
    utxos = list(map(TxUnspent.unspent_from_object, unspents))
    return utxos

def validate_address(address):
    return convert.is_valid(address)

async def build_tx(amount, recipient, memo, fee_rate, sender, network , client_url, key):
    """Build transcation

    :param amount: amount of BTC to transfer
    :type amount: int
    :param recipient: destination address
    :type recipient: str
    :param memo: optional memo for transaction
    :type memo: str
    :param fee_rate: fee rates for transaction
    :type fee_rate: int
    :param sender: sender's address
    :type sender: str
    :param network: testnet or mainnet
    :type network: str
    :param client_url: The haskoin API url
    :type client_url: str
    :param key: bitcash object
    :type key: object
    :returns: transaction
    """
    try:
        utxos = await scan_UTXOs(client_url, sender)
        if len(utxos) == 0:
            raise Exception("No utxos to send")

        balance = (await haskoin_api.get_account(client_url , sender)).confirmed

        if not balance:
            raise Exception("No BCH balance found")

        if not validate_address(recipient):
            raise Exception('Invalid address')

        outputs = [(recipient, amount, 'bch')]

        tx = key.create_transaction(outputs=outputs, fee=int(fee_rate), message=memo, unspents=key.get_unspents(), leftover=sender)
        return tx

    except Exception as err:
        raise Exception(str(err))


import pytest
from xchainpy_bitcoincash import utils

class TestBitcoinCashUtils:
    memo = 'SWAP:THOR.RUNE'

    def test_get_right_vault_fee(self):
        assert utils.calc_fee(fee_rate=10, memo=self.memo) == 1030

    def test_get_normal_fee(self):
        assert utils.calc_fee(fee_rate=10) == 780


from typing import List
from . import utils
from .models.api_types import AddressBalance, Block, Transaction, TransactionInput, TransactionOutput, TxUnspent
import http3
import json


async def get_account(client_url: str, address: str) -> AddressBalance:
    """Get account from address

    :param client_url: The haskoin API url
    :type client_url: str
    :param address: The BCH address
    :type address: str
    :returns: AddressBalance
    """
    try:
        api_url = f'{client_url}/address/{address}/balance'

        client = http3.AsyncClient()
        response = await client.get(api_url)

        if response.status_code == 200:
            balance_response = json.loads(response.content.decode('utf-8'))
            if "error" in balance_response:
                raise Exception(
                    f'Error is : {balance_response["error"]}\nMessage is : {balance_response["message"]}')
            else:
                result = AddressBalance(balance_response['received'], balance_response['utxo'], balance_response['address'],
                                        balance_response['txs'], balance_response['unconfirmed'], balance_response['confirmed'])
                return result
        else:
            return None
    except Exception as err:
        raise Exception(str(err))


async def get_transaction(client_url: str, tx_id: str) -> Transaction:
    """Get transaction by hash

    :param client_url: The haskoin API url
    :type client_url: str
    :param tx_id: The transaction id
    :type tx_id: str
    :returns: Transaction info
    :raises: 'failed to query transaction by a given hash' if failed to query transaction by a given hash
    """
    try:
        api_url = f'{client_url}/transaction/{tx_id}'

        client = http3.AsyncClient()
        response = await client.get(api_url)

        if response.status_code == 200:
            balance_response = json.loads(response.content.decode('utf-8'))

            inputs = []
            for i in balance_response['inputs']:
                inputs.append(TransactionInput(i["pkscript"], i['value'], i['address'], i['witness'],
                                               i['sequence'], i['output'], i['sigscript'], i['coinbase'], i['txid']))
            outputs = []
            for i in balance_response['outputs']:
                outputs.append(TransactionOutput(
                    i["spent"], i['pkscript'], i['value'], i['address'], i['spender']))

            result = Transaction(balance_response["time"], balance_response["size"], inputs, balance_response["weight"], balance_response["fee"], balance_response["locktime"],
                                 balance_response["block"], outputs, balance_response["version"], balance_response["deleted"], balance_response["rbf"], balance_response["txid"])

            return result
        else:
            raise Exception('failed to query transaction by a given hash')
    except Exception as err:
        raise Exception(str(err))


async def get_suggested_tx_fee():
    """Get suggested fee amount for Bitcoin cash. (fee per byte)

    Note: Haskcoin does not provide fee rate related data
    So use Bitgo API for fee estimation
    Refer: https://app.bitgo.com/docs/#operation/v2.tx.getfeeestimate

    :returns: The Bitcoin cash stats
    """
    try:
        api_url = 'https://app.bitgo.com/api/v2/bch/tx/fee'

        client = http3.AsyncClient()
        response = await client.get(api_url)

        if response.status_code == 200:
            fee_response = json.loads(response.content.decode('utf-8'))
            return fee_response["feePerKb"] / 1000 # feePerKb to feePerByte
        else:
            fee_response = json.loads(response.content.decode('utf-8'))
            if "error" in fee_response:
                raise Exception(f'Error is : {fee_response["error"]}')

            return utils.DEFAULT_SUGGESTED_TRANSACTION_FEE
    except:
        return utils.DEFAULT_SUGGESTED_TRANSACTION_FEE

async def get_unspent_transactions(client_url , address) -> List[TxUnspent]:
    """Get unspent transactions

    :param client_url: The haskoin API url
    :type client_url: str
    :param address: The BCH address
    :type address: str
    :returns: The Bitcoin cash stats
    :raises: 'failed to query unspent transactions' if failed to query unspent transactions
    """
    try:
        account = await get_account(client_url , address)

        api_url = f'{client_url}/address/{address}/unspent?limit={account.txs}'

        client = http3.AsyncClient()
        response = await client.get(api_url)

        if response.status_code == 200:
            tx_response = json.loads(response.content.decode('utf-8'))
            result = [TxUnspent(i['pkscript'],i['value'],i['address'],Block(i['block']['height'] , i['block']['position']) ,i['index'],i['txid']) for i in tx_response]
            return result
        else:
            raise Exception('failed to query unspent transactions')
    except Exception as err:
      raise Exception(str(err))

# async def broadcast_tx(client_url, tx_hex):
#     """Broadcast transaction
#     https://sochain.com/api#send-transaction

#     :param client_url: The haskoin API url
#     :type client_url: str
#     :param tx_hex: tranaction hex
#     :type tx_hex: str
#     :returns: Transaction ID
#     """
#     try:
#         api_url = f'{client_url}/transactions'

#         client = http3.AsyncClient()
#         response = await client.post(url=api_url, data=tx_hex)

#         if response.status_code == 200:
#             res = json.loads(response.content.decode('utf-8'))['data']
#             return res['txid']
#         else:
#             return json.loads(response.content.decode('utf-8'))['data']
#     except Exception as err:
#         raise Exception(str(err))


from .utils import get_derive_path
from mnemonic.mnemonic import Mnemonic
import binascii
from bip_utils import Bip32, Bip32Utils

from bitcash import Key, PrivateKeyTestnet

def mnemonic_to_seed(mnemonic, pass_phrase = ''):
    """Convert mnemonic (phrase) to seed

    :param mnemonic: A phrase
    :type mnemonic: str
    :param pass_phrase: A password
    :type pass_phrase: str
    :returns: seed
    """
    mnemo = Mnemonic("english")
    seed = mnemo.to_seed(mnemonic, pass_phrase)
    return seed

def mnemonic_to_private_key(mnemonic,network, pass_phrase = ''):
    """Convert mnemonic (phrase) to a private key

    :param mnemonic: A phrase
    :type mnemonic: str
    :param network: testnet or mainnet
    :type mnemonic: str
    :param pass_phrase: A password
    :type pass_phrase: str
    :returns: private key
    """    
    seed = mnemonic_to_seed(mnemonic, pass_phrase)
    bip32_ctx = Bip32.FromSeed(seed)
    HD_PATH = get_derive_path().testnet if network == "testnet" else get_derive_path().mainnet
    hd_path = HD_PATH[2:]

    bip32_ctx = bip32_ctx.DerivePath(hd_path)
    priv_key = bip32_ctx.PrivateKey().Raw().ToHex()
    return priv_key
    

def private_key_to_address(priv_key, network):
    """Convert a private key to an address

    :param priv_key
    :type public_key: hex
    """
    KeyClass = PrivateKeyTestnet if network == "testnet" else Key
    key = KeyClass.from_hex(priv_key)
    return key.address


import bitcash
from bitcash.network.meta import Unspent


class AddressBalance:
    def __init__(self , received , utxo , address , txs , unconfirmed , confirmed):
        self._received = received
        self._utxo = utxo
        self._address = address
        self._txs = txs
        self._unconfirmed = unconfirmed
        self._confirmed = confirmed

    @property
    def received(self):
        return self._received

    @received.setter
    def received(self, received):
        self._received = received

    @property
    def utxo(self):
        return self._utxo

    @utxo.setter
    def utxo(self, utxo):
        self._utxo = utxo

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, address):
        self._address = address
    
    @property
    def txs(self):
        return self._txs

    @txs.setter
    def txs(self, txs):
        self._txs = txs
        
    @property
    def unconfirmed(self):
        return self._unconfirmed

    @unconfirmed.setter
    def unconfirmed(self, unconfirmed):
        self._unconfirmed = unconfirmed

    @property
    def confirmed(self):
        return self._confirmed

    @confirmed.setter
    def confirmed(self, confirmed):
        self._confirmed = confirmed 

class TransactionInput:
    def __init__(self , pkscript , value , address , witness , sequence , output , sigscript , coinbase , txid):
        self._pkscript = pkscript
        self._value = value
        self._address = address
        self._witness = witness
        self._sequence = sequence
        self._output = output
        self._sigscript = sigscript
        self._coinbase = coinbase
        self._txid = txid
    
    @property
    def pkscript(self):
        return self._pkscript

    @pkscript.setter
    def pkscript(self, pkscript):
        self._pkscript = pkscript 
    
    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value 

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, address):
        self._address = address 

    @property
    def witness(self):
        return self._witness

    @witness.setter
    def witness(self, witness):
        self._witness = witness 

    @property
    def sequence(self):
        return self._sequence

    @sequence.setter
    def sequence(self, sequence):
        self._sequence = sequence 
    
    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, output):
        self._output = output 

    @property
    def sigscript(self):
        return self._sigscript

    @sigscript.setter
    def sigscript(self, sigscript):
        self._sigscript = sigscript 

    @property
    def coinbase(self):
        return self._coinbase

    @coinbase.setter
    def coinbase(self, coinbase):
        self._coinbase = coinbase 

    @property
    def txid(self):
        return self._txid

    @txid.setter
    def txid(self, txid):
        self._txid = txid 

class TransactionOutput:
    def __init__(self ,spent , pkscript , value , address , spender):
        self._spent = spent
        self._pkscript = pkscript
        self._value = value
        self._address = address
        self._spender = spender
    
    @property
    def spent(self):
        return self._spent

    @spent.setter
    def spent(self, spent):
        self._spent = spent 

    @property
    def pkscript(self):
        return self._pkscript

    @pkscript.setter
    def pkscript(self, pkscript):
        self._pkscript = pkscript 
    
    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value 
    
    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, address):
        self._address = address

    @property
    def spender(self):
        return self._spender

    @spender.setter
    def spender(self, spender):
        self._spender = spender


class Transaction:
    def __init__(self , time , size , inputs : TransactionInput , weight , fee , locktime , block , outputs : TransactionOutput , version , deleted , rbf , txid):
        self._time = time
        self._size = size
        self._inputs = inputs
        self._weight = weight
        self._fee = fee
        self._locktime = locktime
        self._block = block
        self._outputs = outputs
        self._version = version
        self._deleted = deleted
        self._rbf = rbf
        self._txid = txid
        
    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, time):
        self._time = time 

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, size):
        self._size = size

    @property
    def inputs(self):
        return self._inputs

    @inputs.setter
    def inputs(self, inputs):
        self._inputs = inputs

    @property
    def weight(self):
        return self._weight

    @weight.setter
    def weight(self, weight):
        self._weight = weight

    @property
    def fee(self):
        return self._fee

    @fee.setter
    def fee(self, fee):
        self._fee = fee

    @property
    def locktime(self):
        return self._locktime

    @locktime.setter
    def locktime(self, locktime):
        self._locktime = locktime
    
    @property
    def block(self):
        return self._block

    @block.setter
    def block(self, block):
        self._block = block

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs

    @property
    def version(self):
        return self._outputs

    @version.setter
    def version(self, version):
        self._version = version

    @property
    def deleted(self):
        return self._deleted

    @deleted.setter
    def deleted(self, deleted):
        self._deleted = deleted

    @property
    def rbf(self):
        return self._rbf

    @rbf.setter
    def rbf(self, rbf):
        self._rbf = rbf

    @property
    def txid(self):
        return self._txid

    @txid.setter
    def txid(self, txid):
        self._txid = txid

class Block:
    def __init__(self , height , position):
        self._height = height
        self._position = position
    
    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position):
        self._position = position

class TxUnspent:
    def __init__(self , pkscript , value , address , block : Block , index , txid):
        self._pkscript = pkscript
        self._value = value
        self._address = address
        self._block = block
        self._index = index
        self._txid = txid

    @classmethod
    def unspent_from_object(cls , unspent):
        txid = unspent.txid
        output_index = unspent.index
        script = bytes.fromhex(unspent.pkscript)
        satoshis = unspent.value
        return Unspent(satoshis , 0, script,txid,output_index)

    @property
    def pkscript(self):
        return self._pkscript

    @pkscript.setter
    def pkscript(self, pkscript):
        self._pkscript = pkscript

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, address):
        self._address = address

    @property
    def block(self):
        return self._block

    @block.setter
    def block(self, block):
        self._block = block

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, index):
        self._index = index

    @property
    def txid(self):
        return self._txid

    @txid.setter
    def txid(self, txid):
        self._txid = txid


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='xchainpy_bitcoincash',
    packages=find_packages(),
    version='0.1',
    license='MIT',
    description='Bitcoincash Module for XChainPy Clients',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='THORChain',
    author_email='',
    url='https://github.com/xchainjs/xchainpy-lib/tree/main/xchainpy/xchainpy_bitcoincash',
    keywords=["THORChain", "XChainpy","XChainpy_bitcoincash"],
    install_requires=required,
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python'
    ],
)